/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2004 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*!
  \file
  \brief Vector copy functions for internal use
  \author Tony Ottosson

  1.3

  2004/10/17 20:48:08
*/

#ifndef __copy_vector_h
#define __copy_vector_h

#include "itconfig.h"
#include "base/binary.h"

#ifndef NO_CBLAS
#include "base/cblas.h"
#endif

using std::complex;
using std::swap;

namespace itpp {

#ifndef DOXYGEN_SHOULD_SKIP_THIS

  /*
    Copy vector x to vector y. Both vectors are of size n
  */
  template<class T> inline
  void copy_vector(const int n, const T *x, T *y)
  {
    for (int i=0; i<n; i++)
      y[i] = x[i];
  }
  template<> inline void copy_vector(const int n, const int *x, int *y) { memcpy(y, x, (unsigned int)n*sizeof(int)); }
  template<> inline void copy_vector(const int n, const short *x, short *y) { memcpy(y, x, (unsigned int)n*sizeof(short)); }
  template<> inline void copy_vector(const int n, const bin *x, bin *y) { memcpy(y, x, (unsigned int)n*sizeof(bin)); }
  template<> inline void copy_vector(const int n, const float *x, float *y) { memcpy(y, x, (unsigned int)n*sizeof(float)); }
  template<> inline void copy_vector(const int n, const complex<float> *x, complex<float> *y) { memcpy(y, x, (unsigned int)n*sizeof(complex<float>)); }
#ifndef NO_CBLAS
  template<> inline void copy_vector(const int n, const double *x, double *y) { cblas_dcopy(n, x, 1, y, 1); }
  template<> inline void copy_vector(const int n, const complex<double> *x, complex<double> *y) { cblas_zcopy(n, x, 1, y, 1); }
#else
  template<> inline void copy_vector(const int n, const double *x, double *y) { memcpy(y, x, (unsigned int)n*sizeof(double)); }
  template<> inline void copy_vector(const int n, const complex<double> *x, complex<double> *y) { memcpy(y, x, (unsigned int)n*sizeof(complex<double>)); }
#endif

  /*
    Copy vector x to vector y. Both vectors are of size n
    vector x elements are stored linearly with element increament incx
    vector y elements are stored linearly with element increament incx
  */
  template<class T> inline
  void copy_vector(const int n, const T *x, const int incx, T *y, const int incy)
  {
    for (int i=0;i<n; i++)
      y[i*incy] = x[i*incx];
  }
#ifndef NO_CBLAS
  template<> inline void copy_vector(const int n, const double *x, const int incx, double *y, const int incy) { cblas_dcopy(n, x, incx, y, incy); }
  template<> inline void copy_vector(const int n, const complex<double> *x, const int incx, complex<double> *y, const int incy) { cblas_zcopy(n, x, incx, y, incy); }
#endif

  /*
    Swap vector x to vector y. Both vectors are of size n
  */
  template<class T> inline
  void swap_vector(const int n, T *x, T *y)
  {
    T tmp;
    for (int i=0; i<n; i++) {
      tmp = y[i];
      y[i] = x[i];
      x[i] = tmp;
    }
  }
  template<> inline void swap_vector(const int n, int *x, int *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
  template<> inline void swap_vector(const int n, short *x, short *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
  template<> inline void swap_vector(const int n, bin *x, bin *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
  template<> inline void swap_vector(const int n, float *x, float *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
  template<> inline void swap_vector(const int n, complex<float> *x, complex<float> *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
#ifndef NO_CBLAS
  template<> inline void swap_vector(const int n, double *x, double *y) { cblas_dswap(n, x, 1, y, 1); }
  template<> inline void swap_vector(const int n, complex<double> *x, complex<double> *y) { cblas_zswap(n, x, 1, y, 1); }
#else
  template<> inline void swap_vector(const int n, double *x, double *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
  template<> inline void swap_vector(const int n, complex<double> *x, complex<double> *y) { for (int i=0; i<n; i++) swap(x[i], y[i]); }
#endif

  /*
    Swap vector x to vector y. Both vectors are of size n
    vector x elements are stored linearly with element increament incx
    vector y elements are stored linearly with element increament incx
  */
  template<class T> inline
  void swap_vector(const int n, T *x, const int incx, T *y, const int incy)
  {
    T tmp;
    for (int i=0; i<n; i++) {
      tmp = y[i*incy];
      y[i*incy] = x[i*incx];
      x[i*incx] = tmp;
    }
  }
  template<> inline void swap_vector(const int n, int *x, const int incx, int *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
  template<> inline void swap_vector(const int n, short *x, const int incx, short *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
  template<> inline void swap_vector(const int n, bin *x, const int incx, bin *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
  template<> inline void swap_vector(const int n, float *x, const int incx, float *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
  template<> inline void swap_vector(const int n, complex<float> *x, const int incx, complex<float> *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
#ifndef NO_CBLAS
  template<> inline void swap_vector(const int n, double *x, const int incx, double *y, const int incy) { cblas_dswap(n, x, incx, y, incy); }
  template<> inline void swap_vector(const int n, complex<double> *x, const int incx, complex<double> *y, const int incy) { cblas_zswap(n, x, incx, y, incy); }
#else
  template<> inline void swap_vector(const int n, double *x, const int incx, double *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
  template<> inline void swap_vector(const int n, complex<double> *x, const int incx, complex<double> *y, const int incy) { for (int i=0; i<n; i++) swap(x[i*incx], y[i*incy]); }
#endif

#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} //namespace itpp

#endif // __copy_vector_h
