/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2004 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*! 
  \file 
  \brief Implementation of Fourier, Hadamard, Walsh-Hadamard, and 2D Hadamard transforms.
  \author Tony Ottosson and Simon Wood 

  This file uses the FFTW package version 3.0.1 that is distributed under the GNU GPL licence. 
  For details see http://www.fftw.org.

  1.17

  2004/09/30 11:53:30
*/

#include "base/matfunc.h"
#include "base/transforms.h"
#include "base/elmatfunc.h"
#include "fftw3.h"
#include <iostream>
#include <cmath>

namespace itpp { 

  //! Local help function
  template <class T>
  void bitrv(Vec<T> &out) {
    int i,j,N1,K,N=out.size();
    T TEMP;

    j=0;
    N1=N-1;
    for (i=0;i<N1;i++) {
      if (i<j) {
	TEMP=out[j];
	out[j]=out[i];
	out[i]=TEMP;
      }
      K=N/2;
      while (K<=j) {
	j=j-K;
	K=K/2;
      }
      j=j+K;
    }
  }

  // --------------- interface to the FFTW routines ------------------------------------------------

#ifndef NO_FFTW

  void fft(const cvec &in, cvec &out)
  {
    static int N;
    out.set_size(in.size(), false);
    
    static fftw_plan p = NULL;

    if (N != in.size()) {
      if (p != NULL)
	fftw_destroy_plan(p);// destroy the previous plan
      // create a new plan
      p = fftw_plan_dft_1d(in.length(),(fftw_complex *)in._data(),(fftw_complex *)out._data(),
			   FFTW_FORWARD, FFTW_ESTIMATE);
      N = in.length();
    }

    //Compute FFT using the GURU FFTW interface
    fftw_execute_dft(p,(fftw_complex *)in._data(),(fftw_complex *)out._data());

  }

  cvec fft(const cvec &in)
  { 
    cvec out(in.size());
    fft(in, out); 
    return out;
  }

  cvec fft(const cvec &in, int N)
  { 
    cvec out(in.size());
    fft(zero_pad(in, N), out); 
    return out;
  }

  void ifft(const cvec &in, cvec &out)
  {
    static int N=0;
    static double inv_N;
    out.set_size(in.size(), false);

    static fftw_plan p = NULL;

    if (N != in.size()) {
      if (p != NULL)
	fftw_destroy_plan(p);// destroy the previous plan
      // create a new plan
      p = fftw_plan_dft_1d(in.length(),(fftw_complex *)in._data(),(fftw_complex *)out._data(),
			   FFTW_BACKWARD, FFTW_ESTIMATE);
      N = in.length();
      inv_N = 1.0/N;
    }

    //Compute IFFT using the GURU FFTW interface
    fftw_execute_dft(p,(fftw_complex *)in._data(),(fftw_complex *)out._data());

    // Scale output
    out *= inv_N;
  }

  cvec ifft(const cvec &in)
  { 
    cvec out(in.size());
    ifft(in, out); 
    return out;
  }

  cvec ifft(const cvec &in, int N)
  { 
    cvec out(in.size());
    ifft(zero_pad(in, N), out); 
    return out;
  }

  void fft_real(const vec &in, cvec &out)
  {
    static int N;
    static fftw_plan p = NULL;
    //out.set_size(in.size(), false);
    out.set_size(in.size()/2+1, false);

    if( N != in.size() )
      {
	if( p!= NULL)
	  fftw_destroy_plan(p); //destroy the previous plan

	// create a new plan
	p = fftw_plan_dft_r2c_1d(in.length(),(double *)in._data(),(fftw_complex *)out._data(),
				 FFTW_ESTIMATE);
	N = in.size();
      }

    //Compute FFT using the GURU FFTW interface
    fftw_execute_dft_r2c(p,(double *)in._data(),(fftw_complex *)out._data());

    //The real FFT does not compute the 2nd half of FFT points because it is redundant
    //to the 1st half. However we want all of the data so we fill it in. This is consistent 
    //with Matlab's functionality
    //int istart = static_cast<int>(std::ceil(in.length()/2.0) + 1) -1;
    //int iend = in.length()-1;
    //int idelta = iend - istart + 1;
    //out.set_subvector(istart,iend,reverse(conj(out(1,idelta))));

  }

  cvec fft_real(const vec &in)
  { 
    cvec out;
    fft_real(in, out);
    return out;
  }

  void ifft_real(const cvec &in, vec &out)
  {
    static int N;
    static double inv_N;

    out.set_size((in.size()-1)*2, false); // always assume an even fft size/Thomas
    static fftw_plan p = NULL;

    if( N != out.size() )
      {
	if( p!=NULL) 
	  fftw_destroy_plan(p); //destroy the previous plan

	//Create a new plan
	p = fftw_plan_dft_c2r_1d(out.size(),(fftw_complex *)in._data(), (double *)out._data(), FFTW_ESTIMATE);
	N = out.size();
	inv_N = 1.0/N;
      }

    //Compute IFFT using the GURU FFTW interface
    fftw_execute_dft_c2r(p,(fftw_complex *)in._data(),(double *)out._data());

    out *= inv_N;
  }

  vec ifft_real(const cvec &in)
  { 
    vec out;
    ifft_real(in, out);
    return out;
  }

#endif //NO_FFTW
  // ---------- FFTW end ---------------------------------------

  template <class T>
  Vec<T> dht(const Vec<T> &v)
  {
    Vec<T> ret(v.size());
    dht(v, ret);
    return ret;
  }

  template <class T>
  void dht(const Vec<T> &vin, Vec<T> &vout)
  {
    T t;
    int m,N,l,k,j,ib,i;

    N = vin.size();
    vout.set_size(N);
  
    m = needed_bits(N);
    it_assert1((1<<m)==N, "dht: The vector size must be a power of two!");

    // This step is separated because it copies vin to vout
    for (ib=0; ib<N; ib+=2) {
      vout(ib) = vin(ib) + vin(ib+1);
      vout(ib+1) = vin(ib) - vin(ib+1);
    }
    N /= 2;
    
    l = 2;    
    for (i=1; i<m; i++) {
      N /= 2;
      ib = 0;
      for (k=0; k<N; k++) {
	for (j=0; j<l; j++) {
	  t = vout(ib+j);
	  vout(ib+j) += vout(ib+j+l);
	  vout(ib+j+l) = t - vout(ib+j+l);
	}
	ib += 2*l;
      }
      l *= 2;
    }

    vout /= T(std::sqrt(double(vin.size())));
  }

  template <class T>
  void self_dht(Vec<T> &v)
  {
    T t;
    int m,N,l=1,k,j,ib,i;

    N = v.size();
  
    m = needed_bits(N);
    it_assert1((1<<m)==N, "dht: The vector size must be a power of two!");

    for (i=0; i<m; i++) {
      N /= 2;
      ib = 0;
      for (k=0; k<N; k++) {
	for (j=0; j<l; j++) {
	  t = v(ib+j);
	  v(ib+j) += v(ib+j+l);
	  v(ib+j+l) = t - v(ib+j+l);
	}
	ib += 2*l;
      }
      l *= 2;
    }

    v /= T(std::sqrt(double(v.size())));
  }

  template <class T>
  Vec<T> dwht(const Vec<T> &v)
  {
    Vec<T> ret(v.size());

    dwht(v, ret);

    return ret;
  }

  template <class T>
  void dwht(const Vec<T> &vin, Vec<T> &vout)
  {
    dht(vin, vout);
    bitrv(vout);
  }


  template <class T>
  void self_dwht(Vec<T> &v)
  {
    self_dht(v);
    bitrv(v);
  }

  template <class T>
  Mat<T> dht2(const Mat<T> &m)
  {
    Mat<T> ret(m.rows(), m.cols());
    Vec<T> v;
    int i;
	
    for (i=0; i<m.rows(); i++) {
      v = m.get_row(i);
      self_dht(v);
      ret.set_row(i, v);
    }
    for (i=0; i<m.cols(); i++) {
      v = ret.get_col(i);
      self_dht(v);
      ret.set_col(i, v);
    }

    return transpose(ret);
  }

  template <class T>
  Mat<T> dwht2(const Mat<T> &m)
  {
    Mat<T> ret(m.rows(), m.cols());
    Vec<T> v;
    int i;
	
    for (i=0; i<m.rows(); i++) {
      v = m.get_row(i);
      self_dwht(v);
      ret.set_row(i, v);
    }
    for (i=0; i<m.cols(); i++) {
      v = ret.get_col(i);
      self_dwht(v);
      ret.set_col(i, v);
    }

    return transpose(ret);
  }

  /////////////////////////////
  //
  //  template instantiation
  //
  /////////////////////////////

  //! Template instantiation of dht
  template vec dht(const vec &v);
  //! Template instantiation of dht
  template cvec dht(const cvec &v);

  //! Template instantiation of dht
  template void dht(const vec &vin, vec &vout);
  //! Template instantiation of dht
  template void dht(const cvec &vin, cvec &vout);

  //! Template instantiation of self_dht
  template void self_dht(vec &v);
  //! Template instantiation of self_dht
  template void self_dht(cvec &v);

  //! Template instantiation of dwht
  template vec dwht(const vec &v);
  //! Template instantiation of dwht
  template cvec dwht(const cvec &v);

  //! Template instantiation of dwht
  template void dwht(const vec &vin, vec &vout);
  //! Template instantiation of dwht
  template void dwht(const cvec &vin, cvec &vout);

  //! Template instantiation of self_dwht
  template void self_dwht(vec &v);
  //! Template instantiation of self_dwht
  template void self_dwht(cvec &v);

  //! Template instantiation of dht2
  template mat  dht2(const mat &m);
  //! Template instantiation of dht2
  template cmat dht2(const cmat &m);

  //! Template instantiation of dht2
  template mat  dwht2(const mat &m);
  //! Template instantiation of dht2
  template cmat dwht2(const cmat &m);

} //namespace itpp
