/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2003 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*!
  \file 
  \brief Implementation of a Matlab Engine Class
  \author

  1.3

  2003/05/22 08:55:20
*/

#include "matlab_engine.h"

namespace itpp {

  Matlab_Engine::Matlab_Engine(const char *start_command){
    if((e = engOpen(start_command)) == NULL)
      it_error("Unable to open the Matlab engine!");
  }

  Matlab_Engine::~Matlab_Engine(){
    if(engClose(e))
      it_error("Unable to close the Matlab engine!");  
  }

  void Matlab_Engine::put(const vec &v, const char *name){
    const int N = v.length();
    mxArray *T;
    if((T = mxCreateDoubleMatrix(N, 1, mxREAL)) == NULL)
      it_error("Unable to allocate Matlab vector. Out of memory?");
    mxSetName(T, name);
    double *pt = (double*) mxGetPr(T);
    for(int k=0; k<N; *pt++ = v(k++));
    if(engPutArray(e, T))
      it_error("Unable to put it++ vector to Matlab workspace!");
  }

  void Matlab_Engine::put(const cvec &v, const char *name){
    const int N = v.length();
    mxArray *T;
    if((T = mxCreateDoubleMatrix(N, 1, mxCOMPLEX)) == NULL)
      it_error("Unable to allocate Matlab vector. Out of memory?");
    mxSetName(T, name);
    double *pt_r = (double*) mxGetPr(T);
    double *pt_i = (double*) mxGetPi(T);
    for(int k=0; k<N; k++){
      *pt_r++ = real(v(k));
      *pt_i++ = imag(v(k));
    }
    if(engPutArray(e, T))
      it_error("Unable to put it++ vector to Matlab workspace!");
  }

  void Matlab_Engine::put(const mat &m, const char *name){
    const int rows = m.rows();
    const int cols = m.cols();
    mxArray *T;
    if((T = mxCreateDoubleMatrix(rows, cols, mxREAL)) == NULL)
      it_error("Unable to allocate Matlab matrix. Out of memory?");
    mxSetName(T, name);
    double *pt = (double*) mxGetPr(T);
    for(int row=0; row<rows; row++)
      for(int col=0; col<cols; *pt++ = m(row, col++));
    if(engPutArray(e, T))
      it_error("Unable to put it++ matrix to Matlab workspace!");
  }

  void Matlab_Engine::put(const cmat &m, const char *name){
    const int rows = m.rows();
    const int cols = m.cols();
    mxArray *T;
    if((T = mxCreateDoubleMatrix(rows, cols, mxCOMPLEX)) == NULL)
      it_error("Unable to allocate Matlab matrix. Out of memory?");
    mxSetName(T, name);
    double *pt_r = (double*) mxGetPr(T);
    double *pt_i = (double*) mxGetPi(T);
    for(int row=0; row<rows; row++){ 
      for(int col=0; col<cols; col++){
	*pt_r++ = real(m(row, col));
	*pt_i++ = imag(m(row, col));
      }
    }
    if(engPutArray(e, T))
      it_error("Unable to put it++ matrix to Matlab workspace!");
  }

  void Matlab_Engine::get(vec &v, const char *name){
    mxArray *T;
    if((T = engGetArray(e, name)) == NULL)
      it_error("No variable with the given name exists in the Matlab workspace!");  
    else{
      const int *dims = mxGetDimensions(T);
      const int ndims = mxGetNumberOfDimensions(T);
      if((ndims>2) || !((dims[0]==1)||(dims[1]==1)))
	it_error("The requested variable is not a vector!");    
      if(mxIsComplex(T))
	it_error("The requested variable is complex-valued!");    
      double *pt = (double*) mxGetPr(T);
      const int N = mxGetNumberOfElements(T);
      v.set_size(N, false);
      for(int k=0; k<N; v(k++) = *pt++);
    }
  }

  void Matlab_Engine::get(cvec &v, const char *name){
    mxArray *T;
    if((T = engGetArray(e, name)) == NULL)
      it_error("No variable with the given name exists in the Matlab workspace!");  
    else{
      const int *dims = mxGetDimensions(T);
      const int ndims = mxGetNumberOfDimensions(T);
      if((ndims>2) || !((dims[0]==1)||(dims[1]==1)))
	it_error("The requested variable is not a vector!");    
      if(!mxIsComplex(T))
	it_error("The requested variable is real-valued!");    
      double *pt_r = (double*) mxGetPr(T);
      double *pt_i = (double*) mxGetPi(T);
      const int N = mxGetNumberOfElements(T);
      v.set_size(N, false);
      if(mxIsComplex(T)) // Copy both real and imaginary part.
	for(int k=0; k<N; k++){
	  complex_double value(*pt_r++, *pt_i++);
	  v(k) = value;
	}
      else // Copy only the real part.
	for(int k=0; k<N; k++){
	  complex_double value(*pt_r++, 0);
	  v(k) = value;
	}         
    }
  }

  void Matlab_Engine::get(mat &m, const char *name){
    mxArray *T;
    if((T = engGetArray(e, name)) == NULL)
      it_error("No variable with the given name exists in the Matlab workspace!");
    else{
      if(mxGetNumberOfDimensions(T)==2){
	double *pt = (double*) mxGetPr(T);
	if(mxIsComplex(T))
	  it_error("The requested variable is complex-valued!");
	const int *dims = mxGetDimensions(T);
	m.set_size(dims[0], dims[1], false);
	for(int row=0; row<dims[0]; row++)
	  for(int col=0; col<dims[1]; m(row, col++)=*pt++);      
      }
      else
	it_error("The requested variable is not a matrix!");
    }
  }

  void Matlab_Engine::get(cmat &m, const char *name){
    mxArray *T;
    if((T = engGetArray(e, name)) == NULL)
      it_error("No variable with the given name exists in the Matlab workspace!");
    else{
      if(mxGetNumberOfDimensions(T)==2){
	double *pt_r = (double*) mxGetPr(T);
	double *pt_i = (double*) mxGetPi(T);   
	const int *dims = mxGetDimensions(T);      
	m.set_size(dims[0], dims[1], false);
	if(mxIsComplex(T)) // Copy both real and imaginary part.
	  for(int row=0; row<dims[0]; row++){ 
	    for(int col=0; col<dims[1]; col++){
	      complex_double value(*pt_r++, *pt_i++);
	      m(row, col) = value;
	    }
	  }          
	else // Copy only the real part.
	  for(int row=0; row<dims[0]; row++){ 
	    for(int col=0; col<dims[1]; col++){
	      complex_double value(*pt_r++, 0);
	      m(row, col) = value;
	    }
	  }      	
      }
      else
	it_error("The requested variable is not a matrix!");
    }
  }

  void Matlab_Engine::put(const ivec &v, const char *name){
    put(to_vec(v), name);
  }

  void Matlab_Engine::put(const bvec &v, const char *name){
    put(to_vec(v), name);
  }

  void Matlab_Engine::put(const imat &m, const char *name){
    put(to_mat(m), name);
  }

  void Matlab_Engine::put(const bmat &m, const char *name){
    put(to_mat(m), name);
  }

  vec Matlab_Engine::get_vec(const char *name){
    vec v;
    get(v, name);
    return v;
  }

  cvec Matlab_Engine::get_cvec(const char *name){
    cvec v;
    get(v, name);
    return v;
  }

  mat Matlab_Engine::get_mat(const char *name){
    mat m;
    get(m, name);
    return m;
  }

  cmat Matlab_Engine::get_cmat(const char *name){
    cmat m;
    get(m, name);
    return m;
  }

  void Matlab_Engine::command(const char *c){
    engEvalString(e, c);
  }

} //namespace itpp 
